/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.builtu.mime;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.RandomAccessFile;
import java.io.Reader;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.builtu.util.commutil.StringUtil;


/**
 *
 * @author zipp
 */
public class MimeUtil {
     private static final Logger logger = Logger.getLogger(MimeUtil.class);

	private static final String [] magicMimeFileLocations = {/*"/etc/magic.mime", "/usr/share/file/magic.mime"*/};

	public static final String UNKNOWN_MIME_TYPE="application/x-unknown-mime-type";


	// the native byte order of the underlying OS. "BIG" or "little" Endian
	private static ByteOrder nativeByteOrder = ByteOrder.nativeOrder();
    // key - file extension,  value = mime Type
    private static Map<String, String> mimeTypes;
    // key - mime Type,  value = Collection<String> of file ext
    private static Map<String, List<String>> fileExt;

    private static ArrayList<MagicMimeEntry> mMagicMimeEntries = new ArrayList<MagicMimeEntry>();

    // Get the native byte order of the OS on which you are running. It will be either
    // Big or little endiun
	public static ByteOrder getNativeOrder() {
		return MimeUtil.nativeByteOrder;
	}

    public static String getFileExtOfMimeType(String mimeType){
        List<String> values = fileExt.get(mimeType);
        if(values == null)
            return null;
        return values.get(0);
    }

    // Always use the magic.mime detection. Do not use the file extension
	public static String getMimeType(File file) {
		String mimeType = null;
		try {
			mimeType = MimeUtil.getMagicMimeType(file);
		}catch(Exception e) {
            logger.warn(e.getMessage());
        }finally {
			if(mimeType == null) {
				mimeType = UNKNOWN_MIME_TYPE;
			}
		}
		return mimeType;
	}
	/** use the magic.mime detection.
     * @see #UNKNOWN_MIME_TYPE
     * @return mime type of content or <code>"application/x-unknown-mime-type"</code>
     * @param content binary content
     * */
	public static String getMimeType(byte[] content) {
		String mimeType = null;
		try {
			mimeType = MimeUtil.getMagicMimeType(content);
		}catch(Exception e) {
            logger.warn(e.getMessage());
		}finally {
			if(mimeType == null) {
				mimeType = UNKNOWN_MIME_TYPE;
			}
		}
		return mimeType;
	}

	/**
	 * Determin the mime type of a file either from its name using
	 * file extension mapping and if that fails use the magig.mime rules.
	 * This requires that the name you pass is the FULL path to the file
	 * You are trying to detect or the magic.mime rule detection will fail
	 * as it will be unable to construct a File object without the path.
	 */
	public static String getMimeType(String fname) {
		String mimeType;
		if(fname == null || "".equals(fname.trim())){
			fname = "";
		}
		// Get the file extension
		String ext = MimeUtil.getFileExtension(fname);
		if(ext == null || "".equals(ext.trim())) {
			// Lets sniff this as we cannot tell from the file extention
			mimeType = MimeUtil.getMimeType(new File(fname));
		} else {
			mimeType = getMimeTypeByExt(ext);
			if(mimeType == null || "".equals(mimeType.trim())) {
				// lets sniff it
				mimeType = MimeUtil.getMimeType(new File(fname));
			}
		}
		return mimeType;
	}
	
	/**
	 * Determin the mime type of a file either only by file extension
	 */
	public static String getMimeTypeByName(String fname) {
		// Get the file extension
		final String ext = MimeUtil.getFileExtension(fname);
		if(ext == null || "".equals(ext.trim())) return null;
		return getMimeTypeByExt(ext);
	}
	
	public static String getMimeTypeByExt(final String ext) {
		// First try case sensitive
		String mimeType = (String)mimeTypes.get(ext);
		if(mimeType == null || "".equals(mimeType.trim())) {
			//try again case insensitive (lower case)
			mimeType = (String)mimeTypes.get(ext.toLowerCase());
		}
		return mimeType;
	}

	// This doesn't actually detect a mime type. What it does do is give you a best match
	// for your requirements. Lets say you get the mime type from a file and it returns
	// a comma seperated list of mime types. You want to pass this back to a browser as
	// a particular format i.e. only one of the available formats, so you pass in a list
	// of formats you will accept i.e. from the browser HTTP accept header and this method
	// will try to determine which of the available mime types best suits your needs.
	public static String getPreferedMimeType(String wanted, String canProvide) {

		// The following is typical of what may be specified in the HTTP Accept header.
		// text/xml, application/xml, application/xhtml+xml, text/html;q=0.9, text/plain;q=0.8, video/x-mng, image/png, image/jpeg, image/gif;q=0.2, text/css, */*;q=0.1
		// The quality parameter (q) indicates how well the user agent handles the MIME type.
		// A value of 1 indicates the MIME type is understood perfectly,
		// and a value of 0 indicates the MIME type isn't understood at all.
		// The reason the image/gif MIME type contains a quality parameter of 0.2, is to indicate that PNG is preferred over GIF
		// if the server is using content negotiation to deliver either a PNG or a GIF to user agents.
		// Similarly, the text/html quality parameter has been lowered a little,
		// to ensure that the XML MIME types are given in preference if content negotiation is being used to serve an XHTML document.
		if(wanted == null || wanted.trim().length() == 0) {
			wanted = "";
		}
//		String [] wantedArray = wanted.split(",");
		String [] wantedArray = StringUtil.split(wanted, ",");
//		String [] canProvideArray = canProvide.split(",");
		String [] canProvideArray = StringUtil.split(canProvide, ",");
		if(canProvideArray.length == 1) {
			// this is all we have to offer
			return canProvideArray[0].trim();
		}
		// Else lets check the rest.

		// TODO. Improve the matching. We need to check for the QoS part to determine the
		// best match. See note above
        for (String aWantedArray : wantedArray) {
            if (aWantedArray.trim().equals("*/*")) {
                return canProvideArray[0].trim();
            }
            for (String aCanProvideArray : canProvideArray) {
                if (aWantedArray.trim().compareToIgnoreCase(aCanProvideArray.trim()) == 0) {
                    return aWantedArray.trim();
                }
            }
        }
        // Assume a wanted type of '*/*' exists and just return the first canProvide type
		return canProvideArray[0].trim();
	}

	private static void putAll(final Map<String, String> self,
			final Properties p) {
		final Enumeration<?> en = p.keys();
		while (en.hasMoreElements()) {
			final String key = (String) en.nextElement();
			final String val = p.getProperty(key);
			self.put(key, val);
		}
	}

	// Initialise the class in preperation for mime type detection
	static {
		mimeTypes = new HashMap<String, String>();
		// Load the file extension mappings from the internal property file and then
		// from the custom property file if it can be found on the classpath
		try {
			// Load the default supplied mime types
			final Properties p = new Properties();
			p.load(MimeUtil.class.getClassLoader().getResourceAsStream("builtu/mime/mime-types.properties"));
			putAll(mimeTypes, p);
			// Load any classpath provided mime types that either extend or override the default mime types
			InputStream is =  MimeUtil.class.getClassLoader(). getResourceAsStream("mime-types.properties");
			if(is != null) {
				try {
					Properties props = new Properties();
					props.load(is);
					if(props.size() > 0) {
                        if(logger.isInfoEnabled())
                            logger.info("Customisations of Mime Types are available");
                        putAll(mimeTypes, props);
					}
				}finally {
					if(is != null) {
						is.close();
					}
				}
			} else {
                if(logger.isInfoEnabled())
				    logger.info("No Customisation or Mime Types available.");
			}
		} catch (IOException ignore) {
            logger.warn(ignore.getMessage());
        }

		// Now parse and initialise the magic.mime rules
		InputStream is;
		// Parse any customised magic.mime file first. This overrides the unix version
		// as any rules you define will always match before the original.
		is = MimeUtil.class.getClassLoader().getResourceAsStream("magic.mime");
		if(is != null) {
			try {
				parse(new InputStreamReader(is));
			}catch(Exception e) {
                logger.warn("Error loading magic.mime: "+e.getMessage());
				// We just won't have any customisations
			}finally {
				if(is != null) {
					try {
						is.close();
					}catch(Exception e) { // ignore
                        logger.warn(e.getMessage(), e);
                    }
                }
				// Set to null for the next search
				is = null;
			}
		}
		// Parse the UNIX magic(5) magic.mime file.
		// This is used should the mime type not be decernable from the extension
		// i.e. the file has no extension
		try {
			boolean found = false;
            for (String magicMimeFileLocation : magicMimeFileLocations) {
                File f = new File(magicMimeFileLocation);
                if (f.exists()) {
                    is = new FileInputStream(f);
                    found = true;
                    break;
                }
            }
            if (!found) {
				// Use the magic.mime that we ship
				is = MimeUtil.class.getClassLoader().getResourceAsStream("builtu/mime/magic.mime");
			}
			parse(new InputStreamReader(is));
		}catch(Exception e) {
			e.printStackTrace();
		}
		finally {
			if(is != null) {
				try {
					is.close();
				}catch(Exception e) { // ignore
                    logger.warn(e.getMessage(), e);
                }
            }
		}

        fileExt = new HashMap<String, List<String>>();
        for(String ext: mimeTypes.keySet()){
            List<String> list = fileExt.get(ext);
            if(list == null){
                list = new ArrayList<String>();
                fileExt.put(ext, list);
            }
            list.add(mimeTypes.get(ext));
        }

    }

	public static void main(String [] args) throws IOException {
		System.out.println(mMagicMimeEntries);

		File dir = new File(".");
		File [] f = dir.listFiles();
        for (File aF : f) {
            System.out.println("file : " + aF.getCanonicalPath() + " : mimeType : " + MimeUtil.getMimeType(aF.getCanonicalPath()));
        }


        System.out.println(MimeUtil.getMajorComponent("application/abcd"));
		System.out.println(MimeUtil.getMajorComponent("application"));
		System.out.println(MimeUtil.getMajorComponent("/application/abcd"));
		System.out.println(MimeUtil.getMajorComponent(""));
		System.out.println(MimeUtil.getMajorComponent(null));

		System.out.println(MimeUtil.getMinorComponent("application/abcd"));
		System.out.println(MimeUtil.getMinorComponent("application"));
		System.out.println(MimeUtil.getMinorComponent("/application/abcd"));
		System.out.println(MimeUtil.getMinorComponent(""));
		System.out.println(MimeUtil.getMinorComponent(null));

	}

	// Parse the magic.mime file
    private static void parse(Reader r) throws IOException {
        BufferedReader br = new BufferedReader(r);
        String line;
        ArrayList<String> sequence = new ArrayList<String>();

        line = br.readLine();
        while (true) {
             if (line == null) {
                break;
            }
            line = line.trim();
            if (line.length() == 0 || line.charAt(0) == '#') {
            	line = br.readLine();
                continue;
            }
            sequence.add(line);

            // read the following lines until a line does not begin with '>' or EOF
            while(true) {
	            line = br.readLine();
	            if(line == null) {
	            	addEntry(sequence);
	            	sequence.clear();
	            	break;
	            }
	            line = line.trim();
	            if (line.length() == 0 || line.charAt(0) == '#') {
	                continue;
	            }
	            if(line.charAt(0) != '>') {
	            	addEntry(sequence);
	            	sequence.clear();
	            	break;
	            }
	            sequence.add(line);
            }

        }
        if(!sequence.isEmpty()) {
        	addEntry(sequence);
        }
    }

    private static void addEntry(ArrayList<?> aStringArray) {
        try {
			MagicMimeEntry magicEntry = new MagicMimeEntry(aStringArray);
            mMagicMimeEntries.add(magicEntry);
		} catch (InvalidMagicMimeEntryException e) {
			// Continue on but lets print an exception so people can see there is a problem
            logger.warn(e.getMessage());
		}
    }

    private static String getMagicMimeType(File f) throws IOException {
    	if(f.isDirectory()) {
    		return "application/directory";
    	}
        int len = mMagicMimeEntries.size();
        RandomAccessFile raf = null;
        try {
            raf = new RandomAccessFile(f, "r");
            for (int i=0; i < len; i++) {
                MagicMimeEntry me = mMagicMimeEntries.get(i);
                String mtype = me.getMatch(raf);
                if (mtype != null) {
                    return mtype;
                }
            }
        } finally {
            if(raf!=null)
                raf.close();
        }
        return null;
    }
    private static String getMagicMimeType(byte[] content) throws IOException {
        int len = mMagicMimeEntries.size();
        for (int i=0; i < len; i++) {
            MagicMimeEntry me = mMagicMimeEntries.get(i);
            String mtype = me.getMatch(content);
            if (mtype != null) {
                return mtype;
            }
        }
        return null;
    }

    // Utility method to get the major part of a mime type
    // i.e. the bit before the '/' character
    public static String getMajorComponent(String mimeType) {
    	if(mimeType == null) {
    		return "";
    	}
    	int offset = mimeType.indexOf("/");
    	if(offset == -1) {
    		return mimeType;
    	} else {
    		return mimeType.substring(0, offset);
    	}
    }

    // Utility method to get the minor part of a mime type
    // i.e. the bit after the '/' character
    public static String getMinorComponent(String mimeType) {
    	if(mimeType == null) {
    		return "";
    	}
    	int offset = mimeType.indexOf("/");
    	if(offset == -1) {
    		return mimeType;
    	} else {
    		return mimeType.substring(offset+1);
    	}
    }

    // Utility method that gets the extension of a file from its name if it has one
    public static String getFileExtension(File file) {
    	try {
    		return MimeUtil.getFileExtension(file.getCanonicalPath());
    	}catch(IOException e) {
    		return "";
    	}
    }

    // Utility method that gets the extension of a file from its name if it has one
	public static String getFileExtension(String fileName) {
		if (fileName == null || fileName.lastIndexOf(".") < 0) {
			return "";
		}
		String extension = fileName.substring(fileName.lastIndexOf(".") + 1);
		// Could be that the path actually had a '.' in it so lets check
		if(extension.contains(File.separator)) {
			extension = "";
		}
		return extension;
	}

}
