/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.util.commutil;

import java.text.ParseException;
import java.util.Date;

import junit.framework.Assert;

import org.builtu.util.commutil.DateUtil;
import org.junit.Test;


public class TestDateUtil {
	
	@Test public void testDay() {
		final Date day = DateUtil.day(2009, 02, 03);
		assertDate("2009-02-03 0000.000", day);
	}
	
	private void assertDate(final String expected, final Date date) {
		final String formatted = DateUtil.format(date);
		Assert.assertEquals(expected, formatted);
	}
	
	@Test public void test_truncateToDay() {
		final Date day = DateUtil.day(2009, 04, 05);
		final Date someTimePassed = new Date(day.getTime()+1000);
		final Date truncated = DateUtil.truncateToDay(someTimePassed);
		Assert.assertEquals(day, truncated);
	}
	
	@Test public void testOffsetHour() {
		final Date day = DateUtil.day(2009, 02, 03);
		final Date oneHour = new Date(day.getTime()+DateUtil.addHour(1));
		assertDate("2009-02-03 0100.000", oneHour);
		final Date endOfDay = new Date(day.getTime()+DateUtil.addHour(23));
		assertDate("2009-02-03 2300.000", endOfDay);
		final Date nextDay = new Date(day.getTime()+DateUtil.addHour(24));
		assertDate("2009-02-04 0000.000", nextDay);
	}
	
	@Test public void testOffsetDay() {
		final Date day = DateUtil.day(2009, 02, 03);
		final Date oneDay = new Date(day.getTime()+DateUtil.addDay(1));
		assertDate("2009-02-04 0000.000", oneDay);
	}

	@Test public void testFormatMillisecond() throws ParseException {
		final Date day = DateUtil.day(2009, 02, 03);
		Assert.assertEquals("2009-02-03 0000.000", DateUtil.format(day, DateUtil.Precision.Millisecond));
	}

	@Test public void testFormatMinute() throws ParseException {
		final Date day = DateUtil.day(2009, 02, 03);
		Assert.assertEquals("2009-02-03 0000", DateUtil.format(day, DateUtil.Precision.Minute));
	}
	
	@Test public void testFormatHour() throws ParseException {
		final Date day = DateUtil.day(2009, 02, 03);
		Assert.assertEquals("2009-02-03 00", DateUtil.format(day, DateUtil.Precision.Hour));
	}

	@Test public void testFormatDay() throws ParseException {
		final Date day = DateUtil.day(2009, 02, 03);
		Assert.assertEquals("2009-02-03", DateUtil.format(day, DateUtil.Precision.Day));
	}

	@Test public void testParseMillisecond() throws ParseException {
		final Date date = DateUtil.parse("2009-02-04 1248.123", DateUtil.Precision.Millisecond);
		assertDate("2009-02-04 1248.123", date);
	}
	
	@Test(expected=ParseException.class) public void testParseMillisecondErr() throws ParseException {
		DateUtil.parse("2009-02-04 1248.12344334", DateUtil.Precision.Millisecond);
	}

	@Test public void testParseMinute() throws ParseException {
		final Date date = DateUtil.parse("2009-02-04 1248", DateUtil.Precision.Minute);
		assertDate("2009-02-04 1248.000", date);
	}
	
	@Test(expected=ParseException.class) public void testParseMinuteErr() throws ParseException {
		DateUtil.parse("2009-02-04 1248.123", DateUtil.Precision.Minute);
	}

	@Test public void testParseHour() throws ParseException {
		final Date date = DateUtil.parse("2009-02-04 12", DateUtil.Precision.Hour);
		assertDate("2009-02-04 1200.000", date);
	}
	
	@Test(expected=ParseException.class) public void testParseHourErr() throws ParseException {
		DateUtil.parse("2009-02-04 1248", DateUtil.Precision.Hour);
	}

	@Test public void testParseDay() throws ParseException {
		final Date date = DateUtil.parse("2009-02-04", DateUtil.Precision.Day);
		assertDate("2009-02-04 0000.000", date);
	}
	
	@Test(expected=ParseException.class) public void testParseDayErr() throws ParseException {
		DateUtil.parse("2009-02-04 12", DateUtil.Precision.Day);
	}

	@Test public void testParseLazyDay() throws ParseException {
		final Date date = DateUtil.parse("2009-02-04");
		assertDate("2009-02-04 0000.000", date);
	}

	@Test public void testParseLazyHours() throws ParseException {
		final Date date = DateUtil.parse("2009-02-04 12");
		assertDate("2009-02-04 1200.000", date);
	}
	
	@Test public void testParseLazyMinutes() throws ParseException {
		final Date date = DateUtil.parse("2009-02-04 1248");
		assertDate("2009-02-04 1248.000", date);
	}
	
	@Test public void testParseLazyMilliseconds() throws ParseException {
		final Date date = DateUtil.parse("2009-02-04 1248.123");
		assertDate("2009-02-04 1248.123", date);
	}
	
	@Test public void testAddDays() throws ParseException {
		final Date day = DateUtil.day(2010, 02, 28);
		final Date add = DateUtil.addDays(day, 1);
		assertDate("2010-03-01 0000.000", add);
	}

	@Test public void testAddMonths() throws ParseException {
		final Date day = DateUtil.day(2010, 01, 31);
		final Date add = DateUtil.addMonths(day, 1);
		assertDate("2010-02-28 0000.000", add);
	}

	@Test public void testAddHours() throws ParseException {
		final Date day = DateUtil.day(2010, 02, 28);
		final Date add = DateUtil.addHours(day, 1);
		assertDate("2010-02-28 0100.000", add);
	}


}
